/****************************************************************************
 * This file is part of the program                                         *
 * mcmcsalut (Markov Chain Monte Carlo Sequence ALignment UTility)          *
 * for joint sampling alignments and mutation parameter values for two      *
 * given DNA-sequences.                                                     *
 *                                                                          *
 *    Copyright (C) 2000  Dirk Metzler                                      *
 *    e-mail: dmetzler@math.uni-frankfurt.de                                *  
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <iostream>
#include <fstream>
#include <string>
#include <vector>

using namespace std;

#include "substproc.hh"
#include "indelproc.hh"
#include "propose.hh"
#include "eingabe.hh"

void eingabe_parameter::interactive()
    /*  Interaktive Eingabe der eingabe_parameter
     */
{
    cout << "input file for sequence 1         : ";
    cin >> dnadat1;
    cout << "input file for sequence 2         : ";
    cin >> dnadat2;
    cout << "output file for parameter samples : ";
    cin >> param_samples;
    start_align=-1;
    while ( start_align<0 || start_align>3) {
	cout << "please choose number of initial alignment type :\n"
	     << " 0 => alignment given in input files\n"
	     << " 1 => sample alignment for parameters given below\n"
	     << " 2 => most probable alignment for parameters given below\n"
	     << " 3 => alignment given in input files; kept fixed\n";
	cin >> start_align;
    }
    cout << "p(A)  : ";
    cin >> basefreq[A];
    cout << "p(C)  : ";
    cin >> basefreq[C];
    cout << "p(G)  : ";
    cin >> basefreq[G];
    cout << "p(T)  : ";
    cin >> basefreq[T];
    cout << "substitution rate                 : ";
    cin >> subrate;
    cout << "additional rate for A<->G         : ";
    cin >> zus_AG;
    cout << "additional rate for C<->T         : ";
    cin >> zus_CT;
    cout << "insertion rate                    : ";
    cin >> lambda;
    cout << "expected length of fragments      : ";
    cin >> gamma;
    cout << "keep fixed (0=variable, 1=fixed):\n";
    cout << "most (base freqs., mean frag. " << endl;
    cout << "length and ratio of all rates)  : ";
    cin >> fix_subst_indel_ratio;
    if (fix_subst_indel_ratio) {
	fix_GC=true;
	fix_AzuT=true;
	fix_CzuG=true;
	fix_subrate=false;
	fix_TizuTv=true;
	fix_PuzuPy=true;
	fix_lambda=false;
	fix_gamma=true;
    }
    else {
	cout << "    AT/GC ratio                   : ";
	cin >> fix_GC;
	cout << "    A/T ratio                     : ";
	cin >> fix_AzuT;
	cout << "    G/C ratio                     : ";
	cin >> fix_CzuG;
	cout << "    substitution rate             : ";
	cin >> fix_subrate;
	cout << "    ratio of Ti/Tv rates          : ";
	cin >> fix_TizuTv;
	cout << "    ratio of Pu/Py Ti rates       : ";
	cin >> fix_PuzuPy;
	cout << "    indel rate                    : ";
	cin >> fix_lambda;
	cout << "    mean fragnment length         : ";
	cin >> fix_gamma;
    }
    if (!fix_gamma) {
	cout << "expectation of prior for (gamma-1): ";
	cin >> prior_gamma_erw;
    }
    cout << "\nno. of runs ...\n"
	 << "... during burn-in                : ";
    cin >> anz_burn;
    cout << "... after burn-in                 : ";
    cin >> anz_runs;
    cout << "... in each sampling interval     : ";
    cin >> anz_steps;
    cout << "size of resampling range          : ";
    cin >> resamp;
};

void eingabe_parameter::from_file(char * dateiname)
    /*
     * Eingabe der eingabe_parameter aus Datei dateiname 
     */
{
    ifstream inputstr(dateiname);
    if (!inputstr) {
	cerr << "cannot open file " << dateiname << endl;
	exit(EXIT_FAILURE);
    }
    inputstr >> dnadat1 >> dnadat2 >> param_samples >> start_align 
	     >> basefreq[A]>> basefreq[C] >> basefreq[G]>> basefreq[T]
	     >> subrate >> zus_AG >> zus_CT >> lambda >> gamma 
	     >> fix_subst_indel_ratio ;
    if(fix_subst_indel_ratio) {	
	fix_GC=true;
	fix_AzuT=true;
	fix_CzuG=true;
	fix_subrate=false;
	fix_TizuTv=true;
	fix_PuzuPy=true;
	fix_lambda=false;
	fix_gamma=true;
    }
    else 
	inputstr >> fix_GC >> fix_AzuT >> fix_CzuG 
		 >> fix_subrate >> fix_TizuTv >> fix_PuzuPy 
		 >> fix_lambda >> fix_gamma;
    if (!fix_gamma) 
	inputstr >> prior_gamma_erw;
    inputstr >> anz_burn >> anz_runs >> anz_steps 
	     >> resamp ;
    inputstr.close();   
    cout << "input file for sequence 1         : ";
    cout << dnadat1 << endl;
    cout << "input file for sequence 2         : ";
    cout << dnadat2 << endl;
    cout << "output file for parameter samples : ";
    cout << param_samples << endl;
    cout << "initial alignment type            :\n";
    switch (start_align) {
    case 0 : 
      cout << "alignment given in input files\n"; 
      break;
    case 1 : 
      cout << "sample alignment for parameters given below\n"; 
      break;
    case 2 : 
      cout << "most probable alignment for parameters given below\n"; 
      break;
    case 3 : 
      cout << "alignment given in input files; kept fixed\n";
    }
    cout << "initial nucleotide probabilities:\n";
    cout << "p(A)  : ";
    cout << basefreq[A] << endl;
    cout << "p(C)  : ";
    cout << basefreq[C] << endl;
    cout << "p(G)  : ";
    cout << basefreq[G] << endl;
    cout << "p(T)  : ";
    cout << basefreq[T] << endl;
    cout << "substitution rate                 : ";
    cout << subrate << endl;
    cout << "additional rate for A<->G         : ";
    cout << zus_AG << endl;
    cout << "additional rate for C<->T         : ";
    cout << zus_CT << endl;
    cout << "insertion rate                    : ";
    cout << lambda << endl;
    cout << "expected length of fragments      : ";
    cout << gamma << endl;
    cout << "keep fixed (0=variable, 1=fixed):\n";
    cout << "  ratio of all rates              : ";
    cout << fix_subst_indel_ratio << endl;
    cout << "    AT/GC ratio                   : ";
    cout << fix_GC << endl;
    cout << "    A/T ratio                     : ";
    cout << fix_AzuT << endl;
    cout << "    G/C ratio                     : ";
    cout << fix_CzuG << endl;
    cout << "    substitution rate             : ";
    cout << fix_subrate << endl;
    cout << "    ratio of Ti/Tv rates          : ";
    cout << fix_TizuTv << endl;
    cout << "    ratio of Pu/Py Ti rates       : ";
    cout << fix_PuzuPy << endl;
    cout << "    indel rate                    : ";
    cout << fix_lambda << endl;
    cout << "    mean fragnment length         : ";
    cout << fix_gamma << endl;
    if (!fix_gamma) 
	cout << "expectation of prior for (gamma-1): " 
	     << prior_gamma_erw << endl;
    cout << "no. of runs ...\n"
	 << "... during burn-in                : ";
    cout << anz_burn << endl;
    cout << "... after burn-in                 : ";
    cout << anz_runs << endl;
    cout << "... in each sampling interval     : ";
    cout << anz_steps << endl;
    cout << "size of resampling range          : ";
    cout << resamp << endl;
}

substproc eingabe_parameter::substpr() const { 
    /* gibt einen substproc gemaess der eingegebenen Substitutionsraten zurueck
     */
    return substproc(basefreq,subrate,zus_AG,zus_CT);
}

fragment_indelproc eingabe_parameter::indelpr() const {
    /* gibt einen fragment_indelproc gemaess der eingegeben InDelraten zurueck
     */
    return fragment_indelproc(lambda,gamma);
}

vector<bool> eingabe_parameter::substfix() const {
    /* gibt zurueck, welche Substitutionsratenrelationen fixiert gesetzt sind
     */
    
    vector<bool> bs(6);
    bs[0]=fix_GC;
    bs[1]=fix_AzuT;
    bs[2]=fix_CzuG;
    bs[3]=fix_subrate;
    bs[4]=fix_TizuTv;
    bs[5]=fix_PuzuPy;
    return bs;
}

vector<bool> eingabe_parameter::indelfix() const {
    /* gibt zurueck, welche Indelratenrelationen fixiert gesetzt sind
     */
    vector<bool> bs(2);
    bs[0]=fix_lambda;
    bs[1]=fix_gamma;
    return bs;
}
