/****************************************************************************
 * This file is part of the program                                         *
 * mcmcsalut (Markov Chain Monte Carlo Sequence ALignment UTility) salut    *
 * joint sampling alignments and mutation parameter values for two given    *
 * DNA-sequences.                                                           *
 *                                                                          *
 *    Copyright (C) 2000  Dirk Metzler                                      *
 *    e-mail: dmetzler@math.uni-frankfurt.de                                * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <string>
#include <vector>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <cmath>
#include <cstdlib>
#include <climits>  

using namespace std;

#include "substproc.hh"
#include "indelproc.hh"
#include "propose.hh"
#include "eingabe.hh"
#include "fragal.hh"

inline double addlog(double a, double b) {
    // liefert log(exp(a)+exp(b))
    if (a>b) return a+log(1+exp(b-a));
    else return b+log(1+exp(a-b));
}

inline double addlog(double a, double b, double c) {
    // liefert log(exp(a)+exp(b)+exp(c))
    return addlog(addlog(a,b),c);
}

inline double maximum(double a, double b) {
    return ( a>b ? a : b );
}

inline double maximum(double a, double b, double c) {
    return maximum(maximum(a,b),c);
}


//+++++++++++++++++++++++++++++++++++++++++++
//        class alignment_mit_raten         +
//+++++++++++++++++++++++++++++++++++++++++++

// private :

void alignment_mit_raten::set_max_wk() {
    /* Setzt das Alignment auf das wahrscheinlichste fuer die aktuellen 
     * Parameter
     */
    zustand nextstate=ENDE;
    vector<zustand> zust_vekt;
    maxwk_align_graph graph(*this);
    // cout << graph;
    int i=dna_A->laen(), j=dna_B->laen();
    while( i>0 || j>0 ) {
	nextstate=graph.maxwk_schritt(nextstate,i,j);
	switch (nextstate) {
	case PAARUNG : i--; j--; break;
	case A_GAP : j--; break;
	case B_GAP : i--; break;
	default: cerr << "unmoeglicher \"nextstate\" in resample!!!\n";
	}
	zust_vekt.push_back(nextstate);
    }
    int a=-1, b=-1;
    for( i=zust_vekt.size()-1; i>=0; i--) {
	switch(zust_vekt[i]) {
	case PAARUNG : 
	    a++; 
	    b++; 
	    dAB[a]=b;
	    dBA[b]=a; 
	    break;
	case A_GAP : 
	    b++; 
	    dBA[b]=a;
	    break;
	case B_GAP : 
	    a++;
	    dAB[a]=b;
	    break;
	default: 
	    cerr << "unmoeglicher \"nextstate\" in resample!!!\n";
	}
    }
}

double alignment_mit_raten::get_ln_wk(substproc sp, fragment_indelproc fip) const {
    /* liefert den ln der Wkeit des aktuellen Alignments fuer die uebergebenen 
     * Parameter.
     */
    int anz_ueberg[4][5]      // Anzahl der einzelnen Uebergaenge 
	={  {0,0,0,0,0},
	    {0,0,0,0,0},
	    {0,0,0,0,0},
	    {0,0,0,0,0} };
    int anz_paare[4][4]      // Anzahl der verschiedenen Basenpaarungen
	={  {0,0,0,0},
	    {0,0,0,0},
	    {0,0,0,0},
	    {0,0,0,0} };
    int anz_indel_basen[4]={0,0,0,0}; // Anzahl an A,C,G,T bei Indels
    zaehle_ueberg(anz_ueberg, anz_paare, anz_indel_basen,
		  -1,-1,dna_A->laen(),dna_B->laen());
    double summe=0;
    for (int i=0;i<4;i++) {
	for (int j=0;j<4;j++) {
	    summe+=anz_ueberg[i][j]*fip.log_wkeit((zustand) i,(zustand) j);
	    summe+=anz_paare[i][j]*sp.log_wkeit((nukleotid) i,(nukleotid) j);
	}
	summe+=anz_ueberg[i][4]*fip.log_wkeit((zustand) i,(zustand) 4);
	summe+=anz_indel_basen[i]*log(sp.get_basefreqs()[(nukleotid) i]);
    }
    return summe;
}


void alignment_mit_raten::zaehle_ueberg(int anz_ueberg[4][5],
					int anz_paare[4][4],
					int anz_indel_basen[4],
					int von_A,int von_B,int bis_A,int bis_B) const {
    /* Berechnet die Anzahlen der verschiedenen Uebergaenge zwischen den
     * Zustaenden, die Anzahl der verschiedenen Basenpaarungen sowie der nichtalignierten
     * Basen im aktuellen Alignment im angegebenen Bereich 
     * und schreibt die Ergebnisse in die uebergebenen Felder.
     */

    zustand oldstate=get_zust(von_A,von_B), zust;
    int a=von_A, b=von_B;
    for (int i=0;i<4;i++) {
	for (int j=0;j<4;j++) {
	    anz_ueberg[i][j]=0;
	    anz_paare[i][j]=0;
	}
	anz_ueberg[i][4]=0;
	anz_indel_basen[i]=0;
    }
    if ( a!=-1 && b!= -1 && dBA[b]!=a && dAB[a]!=b || 
	 ( bis_A != dna_A->laen() && dAB[bis_A]!=bis_B
	   && bis_B != dna_B->laen() && dBA[bis_B]!=bis_A ) )
	cerr << "Fehler bei den Bereichsgrenzen in alignment_mit_raten::zaehle_ueberg\n"
	     << setw(4) << a << setw(4) << b 
	     <<setw(4) << bis_A << setw(4) << bis_B << endl 
	     << setw(4) << dBA[b] << setw(4) << dAB[a] 
	     <<setw(4) << dna_A->laen() << setw(4) << dna_B->laen() 
	     << endl << endl ;
    //cout << 0 << ' ' << 0 << ' ' << dAB[0] << ' ' << dBA[0] << endl;
    do {
	// cout << a << ' ' << b << ' ' << dAB[a] << ' ' << dBA[b] << endl;
	zust=get_zust(a,b);
	switch (zust) {
	case PAARUNG :
	    anz_paare[dna_A->site(a)][dna_B->site(b)]++;
	    break;
	case START :
	    break;
	case A_GAP :
	    anz_indel_basen[dna_B->site(b)]++;
	    break;
	case B_GAP :
	    anz_indel_basen[dna_A->site(a)]++;
	    break;
	case ENDE :
	    break;
	default:	 
	    cerr << "Inkompatibilitaet zwischen a und b in"
		 << " alignment_mit_raten::zaehle_ueberg(..)!!!\n"
		 << "a=" << a << " dAB[a]=" << dAB[a]
		 << " b="  << b << " dBA[b]=" << dBA[b]<< " " <<endl;
	}
	++(anz_ueberg[oldstate][zust]);
	oldstate=iter_zust(&a,&b);
    } while(zust!=ENDE && a<=bis_A && b<=bis_B );
}

zustand alignment_mit_raten::iter_zust(int* pa,int* pb) const {
    /* Liefert den zu *pa und *pb assoziierten Alignment-Zustand und 
     * setzt (*pa,*pb) einen Zustand weiter.
     */
    zustand zust;
    zust=get_zust(*pa,*pb);
    switch (zust) {
    case START:
    case PAARUNG:  
    case A_GAP: 
    case B_GAP:
	if ( *pa==dna_A->laen()-1 ) {
	    ++*pb;
	    if ( *pb==dna_B->laen() ) 
		++*pa;
	}
	else {
	    if ( *pb==dna_B->laen()-1 )
		++*pa;
	    else {
		if ( dAB[*pa+1]==*pb )
		    ++*pa;
		else {
		    ++*pb;
		    if ( dBA[*pb]!=*pa ) 
			++*pa;
		}
	    }
	}
	break;
    case ENDE:
	break;
    default:
	cerr << *pa << " und " << *pb << " ergeben keinen Zustand"
	     << " in alignment_mit_raten::iter_zust\n";
	exit(1);
    }
    return zust;
}

void alignment_mit_raten::teilsample(int von_A, int von_B, int bis_A, int bis_B) 
{
    /* sample ein neues Teilalignment von (von_A,von_B) 
     * bis (bis_A,bis_B). Dabei bleiben die Zustaende von
     * Start- und Endpunkt erhalten.
     */
    
    zustand nextstate=get_zust(bis_A,bis_B);
    vector<zustand> zust_vekt;
    int a=bis_A, b=bis_B;
    int anz_A_GAP=0,anz_B_GAP=0;
    while (iter_zust(&a,&b)==A_GAP) ++anz_A_GAP;
    if (anz_A_GAP==0) {
	a=bis_A;
	b=bis_B;
	while (iter_zust(&a,&b)==B_GAP) ++anz_B_GAP;
    }
    if ( von_A+1 < bis_A && von_B+1 < bis_B && bis_A >= 0 && bis_B>= 0 ) {
	wk_align_graph graph(*this,von_A,von_B,bis_A,bis_B);
	// cout << graph;
	double ln_wk_alt=get_ln_wk();
	int i=0, j=0;
	switch (nextstate) {
	case ENDE :
	    i=dna_A->laen();
	    j=dna_B->laen();
	    break;
	case PAARUNG :
	    i=bis_A;
	    j=bis_B;
	    break;
	case A_GAP :
	    i=bis_A+1;
	    j=bis_B;
	    break;
	case B_GAP :
	    i=bis_A;
	    j=bis_B+1;
	    break;
	default:
	    cerr << "unmoeglicher \"nextstate\" in teilsample(...)\n";
	}
	//   cout << endl << "A:"<< von_A << '-' << bis_A << ' ' 
	//        << "B:" << von_B << '-' << bis_B<< endl;
	while( i>von_A+1 || j>von_B+1 ) {
	    nextstate=graph.zufallsschritt(nextstate,i,j);
	    switch (nextstate) {
	    case PAARUNG : i--; j--; break;
	    case A_GAP : j--; break;
	    case B_GAP : i--; break;
	    default: cerr << "unmoeglicher \"nextstate\" in resample!!!\n";
	    }
	    if ( i<von_A || j<von_B ) 
		cerr << "Bereich verlassen in  alignment_mit_raten::teilsample(...)\n";
	    zust_vekt.push_back(nextstate);
	    // cout << i << " " << j << " " << nextstate << endl;
	}
	a=von_A;
	b=von_B;
	for( i=zust_vekt.size()-1; i>=0; i--) {
	    switch(zust_vekt[i]) {
	    case PAARUNG : 
		a++; 
		b++;  
		dAB[a]=b;
		dBA[b]=a;
		break;
	    case A_GAP :
		b++;  
		dBA[b]=a;
		break;
	    case B_GAP :
		a++;
		dAB[a]=b; 
		break;
	    case ENDE :
		break;
	    default: 
		cerr << "unmoeglicher \"nextstate\" in teilsample!!!\n";
	    }
	    //cout << a << " " << b << " " <<  setw(10) << zust_vekt[i] 
	    //	 << ' ' << (*dna_A)[a] << ' ' <<  (*dna_B)[b] <<endl;
	}
	for( i=1; i<anz_A_GAP; i++){
	    dBA[bis_B+i]=dBA[bis_B];
	    //cout << "dBA[" << bis_B+i << "]=" << dBA[bis_B] << endl;
	}
	for( i=1; i<anz_B_GAP; i++){
	    dAB[bis_A+i]=dAB[bis_A];
	    //cout <<"dAB[" << bis_A+i << "]=" << dAB[bis_A] << endl;
	}
	if (get_ln_wk()<ln_wk_alt-100){
	    cerr << endl << "Drastische Reduziereung der Wahrscheinlichkeit in teilsample" << endl
		 << get_ln_wk() << ' ' << ln_wk_alt << endl
		 << setw(4) << von_A<< setw(4) << von_B
		 << setw(4) << bis_A<< setw(4) << bis_B<<endl
		 << *this << endl 
		 << graph << endl
		 << "Abbruch in alignment_mit_raten::teilsample(int,int,int,int)\n"
		;
	    exit(1);
	}
    }
}

void alignment_mit_raten::coupled_param_resamp() {
    /* Fuehrt einen Metropolis-Hastings-Schritt fuer die Mutationsraten
     * bedingt auf das aktuelle Alignment durch. Dabei bleiben die Verhaeltnisse
     * zwischen den verschiedenen Raten erhalten. Es entspricht also einem
     * samplen der Zeitspanne zwischen den Sequenzen.
     */
    double t=rand_exp(1.0);
    indel_vorschlag.vorschlag(t);
    subst_vorschlag.vorschlag(t);
    double quotient=exp( get_ln_wk(subst_vorschlag.get_vorlauf(),indel_vorschlag.get_vorlauf()) 
			 - get_ln_wk() ) 
	* indel_vorschlag.quot_prior() * subst_vorschlag.quot_prior()
	* exp(t-1/t)/t ;
    if ( quotient >= 1.0 || (rand()+1.0)/(RAND_MAX+1.0) < quotient ) {
	indel_vorschlag.akzeptiere();
	subst_vorschlag.akzeptiere();
    }
}

void alignment_mit_raten::indel_param_resamp() {
    /* Fuehrt einen Metropolis-Hastings-Schritt fuer die indel-parameter
     * bedingt auf das aktuelle Alignment und aktuelle substparameter durch.
     */
    indel_vorschlag.vorschlag();
    double quotient=exp( get_ln_wk(subst_vorschlag.get_aktuell(),indel_vorschlag.get_vorlauf()) 
			 - get_ln_wk() ) 
	* indel_vorschlag.quot_prior()*indel_vorschlag.uebergangs_quotient() ;
    if ( quotient >= 1.0 || (rand()+1.0)/(RAND_MAX+1.0) < quotient ) 
	indel_vorschlag.akzeptiere();
}

void alignment_mit_raten::subst_param_resamp() {
    /* Fuehrt einen Metropolis-Hastings-Schritt fuer die subst-parameter
     * bedingt auf das aktuelle Alignment und aktuelle indelparameter durch.
     */

    subst_vorschlag.vorschlag();
    double quotient=  exp( get_ln_wk(subst_vorschlag.get_vorlauf(),indel_vorschlag.get_aktuell())
			   - get_ln_wk() ) 
	* subst_vorschlag.quot_prior()*subst_vorschlag.uebergangs_quotient();
    if ( quotient>=1.0 || (rand()+1.0)/(RAND_MAX+1.0)< quotient ) {
	subst_vorschlag.akzeptiere();
    }
}

// public:

alignment_mit_raten::alignment_mit_raten(const dna &d_1,const dna &d_2,
	      const dna_with_gaps &d_1wg,const dna_with_gaps &d_2wg,
	      const eingabe_parameter & ep)
    /* Konstruiert das Alignment gemaess den Eingabeparametern in eg
     */ 
    : indel_vorschlag(ep.indelpr(),ep.indelfix(),ep.prior_gamma_erw),
      subst_vorschlag(ep.substpr(),ep.substfix()),
      dAB(d_1.laen()), dBA(d_2.laen())
{
    int a=-1,b=-1; 
    dna_A=&d_1;
    dna_B=&d_2;
    fix_subst_indel_ratio=ep.fix_subst_indel_ratio;
    switch (ep.start_align) {
    case 0 :
    case 3 :
	if(d_1wg.laen()!=d_2wg.laen()) {
	    cerr << "dna_with_gap's have different lengths!" << endl;
	    exit(1);
	}
	for(int i=0; i<d_1wg.laen(); i++){
	    if(d_1wg[i]=='_') {
		if(d_2wg[i]!='_') {
		    b++;
		    dBA[b]=a;
		}
	    }
	    else {
		if(d_2wg[i]=='_') {
		    a++;
		    dAB[a]=b;
		}
		else {
		    a++;
		    b++;
		    dAB[a]=b;
		    dBA[b]=a;
		} 
	    }
	}
	break;
    case 1 :
	set_max_wk();
	sample();
	break;
    case 2 : 
	set_max_wk();
	break;
    default :
	cerr << "unknown type of initial alignment_mit_raten" << endl;
	exit(EXIT_FAILURE);
    }
}

void alignment_mit_raten::sample()
    /* sample das Alignment
     */
{
    teilsample(-1,-1,dna_A->laen(),dna_B->laen());
    //int a[3][4],anz_matches,anz_mismatches;
    //zaehle_ueberg(a,&anz_matches, &anz_mismatches, 
    //		  0,0,dna_A->laen(),dna_B->laen());
    //   for( int i=0;i<3;i++){
    //     for (int j=0 ; j<4;j++)
    //       cout << setw(4) << a[i][j];
    //     cout << endl;
    //   }
    //   cout << anz_matches << " " << anz_mismatches << endl;
} 

void alignment_mit_raten::teilsample(int von_A, int bis_A) {
    /* sample ein neues Teilalignment von dna_A[von_A] 
     * bis dna_A[bis_A].
     */
    teilsample(von_A,dAB[von_A],bis_A,dAB[bis_A]);
}

void alignment_mit_raten::teilsample(int laenge) {
    /* sample ein neues Teilalignment zu einem zufaelligen Abschnitt der 
     * Laenge laenge. 
     */
    if (laenge<1) cerr << "unsinniger Parameter in alignment_mit_raten::teilsample(...)\n";
    if (laenge>=dna_A->laen() ) sample();
    else {
	int R=rand()%(dna_A->laen()+laenge);
	if ( R >= dna_A->laen() )
	    teilsample( -1,-1, laenge, dAB[laenge]);
	else {
	    if ( R+laenge >= dna_A->laen() ) 
		teilsample( dna_A->laen()-laenge, dAB[dna_A->laen()-laenge],
			    dna_A->laen(),dna_B->laen());
	    else
		teilsample( R,R+laenge);
	}
    }
}

void alignment_mit_raten::param_resamp() {
    /* Fuehrt einen Metropolis-Hastings-Schritt fuer s
     * lambda und mu bedingt auf das aktuelle Alignment durch.
     */
    if (fix_subst_indel_ratio) 
	coupled_param_resamp();
    else {
	indel_param_resamp();
	subst_param_resamp();
    }
}

void alignment_mit_raten::print_params(ostream& ostr) {
    /* gibt s, lambda und ln(W'keit) in den ostr aus
     */
    vector<double> bf=subst_vorschlag.get_aktuell().get_basefreqs();
    ostr << setw(10) << bf[A] << ' ' << setw(8) << bf[C]  << ' '
	 << setw(10) << bf[G] << ' ' << setw(8) << bf[T]  << ' ' 
	 << setw(10) << subst_vorschlag.get_aktuell().get_allgrate() << ' ' 
	 << setw(10) << subst_vorschlag.get_aktuell().get_zusAG() << ' ' 
	 << setw(10) << subst_vorschlag.get_aktuell().get_zusCT() << ' ' 
	 << setw(10) << indel_vorschlag.get_aktuell().get_lambda() << ' ' 
	 << setw(10) << indel_vorschlag.get_aktuell().get_gamma() << ' ';
}

void alignment_mit_raten::print_path(ostream& ostr) {
    /* gibt das Alignment Pfad in ostr aus:
     * d,v,h stehen fuer diagonal, vertikal und horizontal
     */    
    int i=0,a=-1,b=-1;
    string pfad;
    iter_zust(&a,&b);
    zustand zust;
    do {
	zust=get_zust(a,b);
	switch (zust) {
	case PAARUNG :
	    pfad+='d';
	    break;
	case A_GAP :
	    pfad+='h';
	    break;
	case B_GAP :
	    pfad+='v';
	    break;
	case ENDE :
	    break;
	default:
	    cerr << "incompatibility between a and b" <<endl;
	    exit(1);
	}
	i++;
	zust=iter_zust(&a,&b);
    } while(zust!=ENDE);
    ostr << pfad;
}

ostream& operator<<(ostream& ostr, const alignment_mit_raten& algn) {
    int i=0,a=-1,b=-1,anz_matches=0,anz_mismatches=0,anz_A_gaps=0,anz_B_gaps=0;
    string wort_A, wort_B;
    algn.iter_zust(&a,&b);
    zustand zust;
    do {
	zust=algn.get_zust(a,b);
	switch (zust) {
	case PAARUNG :
	    wort_A+=(*algn.dna_A)[a];
	    wort_B+=(*algn.dna_B)[b];
	    if ((*algn.dna_A)[a]==(*algn.dna_B)[b])
		anz_matches++;
	    else
		anz_mismatches++;
	    break;
	case A_GAP :
	    wort_A+='_';
	    wort_B+=(*algn.dna_B)[b];
	    anz_A_gaps++;
	    break;
	case B_GAP :
	    wort_A+=(*algn.dna_A)[a];
	    wort_B+='_';
	    anz_B_gaps++;
	    break;
	case ENDE :
	    break;
	default:
	    ostr << wort_A << endl << wort_B << endl;
	    cerr << "incompatibility between a and b in"
		 << " ostream& operator<<(...)!!!\n"
		 << "a=" << a << " dAB[a]=" << algn.dAB[a]
		 << " b="  << b << " dBA[b]=" << algn.dBA[b]<< " " <<endl;
	    exit(1);
	}
	i++;
	if( i%70==0) {
	    ostr << wort_A << endl << wort_B << endl << endl;
	    wort_A="";
	    wort_B="";
	}
	zust=algn.iter_zust(&a,&b);
    } while(zust!=ENDE);
    ostr << wort_A;
    ostr << endl  << wort_B;
    ostr << endl << algn.subst_vorschlag.get_aktuell();
    ostr << endl << algn.indel_vorschlag.get_aktuell();
    ostr << endl
	 << "logarithmus naturalis of probability of this alignment : "
	 << algn.get_ln_wk() << endl;
//   int az=-1, bz=-1;
//   zustand zust=algn.iter_zust(&az,&bz);
//   while (zust!=ENDE) {
//     zust=algn.iter_zust(&az,&bz);
//     cout << zust << ",";
//   }
    return ostr << endl;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++
//             class align_graph                     +
//++++++++++++++++++++++++++++++++++++++++++++++++++++

//  private :
//  public :

align_graph::align_graph(const alignment_mit_raten& algn,
			 int v_A,int v_B,int b_A,int b_B)
    /* Konstruiere einen Alignment-Graphen fuer das Teilalignment zwischen
     * (von_A,von_B) und (bis_A,bis_B) in algn mit den aktuellen Parametern
     * von algn
     */
    : 
      sp(algn.subst_vorschlag.get_aktuell()),
      fip(algn.indel_vorschlag.get_aktuell()),
      A_gap_kante(b_A-v_A+2,vector<double>(b_B-v_B+1,999999.9)) ,
      B_gap_kante(b_A-v_A+1,vector<double>(b_B-v_B+2,999999.9)) ,
      paar_kante(b_A-v_A+1,vector<double>(b_B-v_B+1,999999.9)) 
{
 
    von_A=v_A; bis_A=b_A; von_B=v_B; bis_B=b_B;
}

align_graph::align_graph(const alignment_mit_raten& algn) :
    sp(algn.subst_vorschlag.get_aktuell()),
    fip(algn.indel_vorschlag.get_aktuell()),
    A_gap_kante(algn.get_dna_A()->laen()+1, 
		vector<double>(algn.get_dna_B()->laen(),999999.9)) ,
    B_gap_kante(algn.get_dna_A()->laen(),
		vector<double>(algn.get_dna_B()->laen()+1,999999.9)) ,
    paar_kante(algn.get_dna_A()->laen(),
	       vector<double>(algn.get_dna_B()->laen(),999999.9)  )
    
{
    /* Konstruiere einen Alignment-Graphen fuer das Alignment
     */
    von_A=-1; bis_A=algn.get_dna_A()->laen();
    von_B=-1; bis_B=algn.get_dna_B()->laen();
}

ostream& operator<<(ostream& ostr, const align_graph& ag) {
    for(int i=0; i<=ag.bis_A-ag.von_A-2; i++) 
	ostr << setw(30) << i+ag.von_A+1;
    ostr<<endl;
    for(int j=0; j<=ag.bis_B-ag.von_B-2; j++) {
	ostr << setw(5) << " ";
	for(int i=0; i<=ag.bis_A-ag.von_A-2; i++) 
	    ostr <<  setw(30) << ag.B_gap_kante[i][j];
	ostr << setw(5) << j+ag.von_B+1;
	for(int i=0; i<=ag.bis_A-ag.von_A-2; i++) 
	    ostr <<  setw(15) << ag.A_gap_kante[i][j]
		 <<  setw(15) << ag.paar_kante[i][j];
	ostr << setw(15) << ag.A_gap_kante[ag.bis_A-ag.von_A-1][j];
	ostr<<endl;
    }
    ostr << setw(5) << " ";
    for(int i=0; i<=ag.bis_A-ag.von_A-2; i++) 
	ostr << setw(30) << ag.B_gap_kante[i][ag.bis_B-ag.von_B-1];
    ostr<<endl;
    return ostr;
}

//+++++++++++++++++++++++++++++++++++++++++++++
//       class   wk_align_graph               +
//+++++++++++++++++++++++++++++++++++++++++++++

//  private :

void wk_align_graph::beschrifte(const alignment_mit_raten& algn) 
    /* Beschriftet die Kanten des align_graph fuer den angegebenen 
     * Bereich gemaess den DNA-Staengen und den Parametern von algn
     */
{  
    sp=algn.subst_vorschlag.get_aktuell();
    fip=algn.indel_vorschlag.get_aktuell();

//     cout << sp << fip << endl;
//     for (int i=0; i<4; i++) {
// 	cout << sp.get_basefreqs()[i] << "    ";
// 	for (int j=0; j<4; j++) 
// 	    cout << exp(sp.log_wkeit((nukleotid)i,(nukleotid)j)) << ' ';
// 	cout << endl;
//     }
//     for (int i=0; i<4; i++) {
// 	for (int j=0; j<4; j++) 
// 	    cout << exp(fip.log_wkeit((zustand)i,(zustand)j)) << ' ';
// 	cout << endl;
//     }
    

    const dna &dna_A=*(algn.get_dna_A());
    const dna &dna_B=*(algn.get_dna_B());
    zustand firststate = algn.get_zust(von_A,von_B), 
	laststate  = algn.get_zust(bis_A,bis_B);
    int v_A=von_A+1,v_B=von_B+1,b_A=0,b_B=0;
    switch (laststate) {
    case ENDE :
	b_A=algn.dna_A->laen()-1;
	b_B=algn.dna_B->laen()-1;
	break;
    case PAARUNG :
	b_A=bis_A-1;
	b_B=bis_B-1;
	break;
    case A_GAP :
	b_A=bis_A;
	b_B=bis_B-1;
	break;
    case B_GAP :
	b_A=bis_A-1;
	b_B=bis_B;
	break;
    default:
	cerr << "komischer laststate in wk_align_graph::beschrifte(...)\n";
	exit(1);
    } 
    // Zur W'keit fuer Substitution verwandte Logarithmen
    int i,j;
    A_gap_kante[0][0]=fip.log_wkeit(firststate,A_GAP)
      +log(sp.get_basefreqs()[nukleo(dna_B[v_B])]);
    B_gap_kante[0][0]=fip.log_wkeit(firststate,B_GAP)
      +log(sp.get_basefreqs()[nukleo(dna_A[v_A])]);
    paar_kante[0][0]=fip.log_wkeit(firststate,PAARUNG)
	+ sp.log_wkeit(nukleo(dna_A[v_A]),nukleo(dna_B[v_B]));
    for(i=1;i<=b_A-v_A;i++) {
	A_gap_kante[i][0]=B_gap_kante[i-1][0]
	    +fip.log_wkeit(B_GAP,A_GAP)
	    +log(sp.get_basefreqs()[nukleo(dna_B[v_B])]);
	B_gap_kante[i][0]=B_gap_kante[i-1][0]
	    +fip.log_wkeit(B_GAP,B_GAP)
	    +log(sp.get_basefreqs()[nukleo(dna_A[i+v_A])]);
	paar_kante[i][0]=B_gap_kante[i-1][0]
	    +fip.log_wkeit(B_GAP,PAARUNG)
	    + sp.log_wkeit(nukleo(dna_A[i+v_A]),nukleo(dna_B[v_B]));
    }
    A_gap_kante[b_A-v_A+1][0]=B_gap_kante[b_A-v_A][0]
	+fip.log_wkeit(B_GAP,A_GAP)
	+log(sp.get_basefreqs()[nukleo(dna_B[v_B])]);
    for(j=1;j<=b_B-v_B;j++) {
	B_gap_kante[0][j]=A_gap_kante[0][j-1]
	    +fip.log_wkeit(A_GAP,B_GAP)
	    +log(sp.get_basefreqs()[nukleo(dna_A[v_A])]);
	A_gap_kante[0][j]=A_gap_kante[0][j-1]
	    +fip.log_wkeit(A_GAP,A_GAP)
	    +log(sp.get_basefreqs()[nukleo(dna_B[j+v_B])]);
	paar_kante[0][j]=A_gap_kante[0][j-1]
	    +fip.log_wkeit(A_GAP,PAARUNG)
	    + sp.log_wkeit(nukleo(dna_A[v_A]),nukleo(dna_B[j+v_B]));
    }
    B_gap_kante[0][b_B-v_B+1]=(A_gap_kante[0][b_B-v_B]
			       +fip.log_wkeit(A_GAP,B_GAP)
			       +log(sp.get_basefreqs()[nukleo(dna_A[v_A])]));
    for(i=1;i<=b_A-v_A;i++) {
	for(j=1;j<=b_B-v_B;j++) {
	    A_gap_kante[i][j]
		=addlog(B_gap_kante[i-1][j] + fip.log_wkeit(B_GAP,A_GAP)
			+ log(sp.get_basefreqs()[nukleo(dna_B[j+v_B])]),
			paar_kante[i-1][j-1] + fip.log_wkeit(PAARUNG,A_GAP)
			+ log(sp.get_basefreqs()[nukleo(dna_B[j+v_B])]),
			A_gap_kante[i][j-1] + fip.log_wkeit(A_GAP,A_GAP)
			+ log(sp.get_basefreqs()[nukleo(dna_B[j+v_B])]));
	    B_gap_kante[i][j]
		=addlog(B_gap_kante[i-1][j] + fip.log_wkeit(B_GAP,B_GAP)
			+ log(sp.get_basefreqs()[nukleo(dna_A[i+v_A])]),
			paar_kante[i-1][j-1] + fip.log_wkeit(PAARUNG,B_GAP)
			+ log(sp.get_basefreqs()[nukleo(dna_A[i+v_A])]),
			A_gap_kante[i][j-1] + fip.log_wkeit(A_GAP,B_GAP)
			+ log(sp.get_basefreqs()[nukleo(dna_A[i+v_A])]));
	    paar_kante[i][j]
		=addlog(B_gap_kante[i-1][j] + fip.log_wkeit(B_GAP,PAARUNG)
			+ sp.log_wkeit(nukleo(dna_A[i+v_A]),nukleo(dna_B[j+v_B])),
			paar_kante[i-1][j-1] + fip.log_wkeit(PAARUNG,PAARUNG)
			+ sp.log_wkeit(nukleo(dna_A[i+v_A]),nukleo(dna_B[j+v_B])),
			A_gap_kante[i][j-1] + fip.log_wkeit(A_GAP,PAARUNG)
			+ sp.log_wkeit(nukleo(dna_A[i+v_A]),nukleo(dna_B[j+v_B])));
	}
    }
    for(i=1;i<=b_A-v_A;i++) 
	B_gap_kante[i][b_B-v_B+1]
	    =addlog(B_gap_kante[i-1][b_B-v_B+1]+fip.log_wkeit(B_GAP,B_GAP)
		    +log(sp.get_basefreqs()[nukleo(dna_A[i+v_A])]),
		    paar_kante[i-1][b_B-v_B]+fip.log_wkeit(PAARUNG,B_GAP)
		    +log(sp.get_basefreqs()[nukleo(dna_A[i+v_A])]),
		    A_gap_kante[i][b_B-v_B]+fip.log_wkeit(A_GAP,B_GAP)
		    +log(sp.get_basefreqs()[nukleo(dna_A[i+v_A])]));
    for(j=1;j<=b_B-v_B;j++)
	A_gap_kante[b_A-v_A+1][j]
	    =addlog(B_gap_kante[b_A-v_A][j]+fip.log_wkeit(B_GAP,A_GAP)
		    +log(sp.get_basefreqs()[nukleo(dna_B[j+v_B])]),
		    paar_kante[b_A-v_A][j-1]+fip.log_wkeit(PAARUNG,A_GAP)
		    +log(sp.get_basefreqs()[nukleo(dna_B[j+v_B])]),
		    A_gap_kante[b_A-v_A+1][j-1]+fip.log_wkeit(A_GAP,A_GAP)
		    +log(sp.get_basefreqs()[nukleo(dna_B[j+v_B])]));
    ln_likelihood=addlog(paar_kante[b_A-v_A][b_B-v_B]+fip.log_wkeit(PAARUNG,ENDE),
			 B_gap_kante[b_A-v_A][b_B-v_B+1]+fip.log_wkeit(B_GAP,ENDE),
			 A_gap_kante[b_A-v_A+1][b_B-v_B]+fip.log_wkeit(A_GAP,ENDE) );
    for (i=0; i<b_A-v_A; i++)
	for (j=0; j<b_B-v_B; j++) {
	    if (A_gap_kante[i][j]>0) {
		cerr << "in wk_align_graph::beschrifte(const alignment_mit_raten& algn):\n"
		     << "A_gap_kante[" << i << "][" << j << "]=" << A_gap_kante[i][j] << endl
		     << b_B-v_B << endl;
		exit(1);
	    }
	    if (B_gap_kante[i][j]>0) {
		cerr << "in wk_align_graph::beschrifte(const alignment_mit_raten& algn):\n"
		     << "B_gap_kante[" << i << "][" << j << "]=" << B_gap_kante[i][j] << endl;
		exit(1);
	    }
	    if (paar_kante[i][j]>0) {
		cerr << "in wk_align_graph::beschrifte(const alignment_mit_raten& algn):\n"
		     << "paar_kante[" << i << "][" << j << "]=" << paar_kante[i][j] << endl
		     << dna_A[v_A+i] << ' ' << dna_B[v_B+j] 
		     << ' ' << sp.log_wkeit(nukleo(dna_A[v_A+i]), nukleo(dna_B[v_B+j])) << endl;
		exit(1);
	    }
	}
}

// public :
wk_align_graph::wk_align_graph(const alignment_mit_raten& algn,
			       int v_A,int v_B,int b_A,int b_B)
    /* Konstruiere einen wk-Alignment-Graphen fuer das Teilalignment zwischen
   * (von_A,von_B) und (bis_A,bis_B) in algn mit den aktuellen Parametern
   * von algn
   */
    :align_graph(algn,v_A,v_B,b_A,b_B)
{
    beschrifte(algn);
}

zustand wk_align_graph::zufallsschritt(zustand nextstep, int a, int b) const
    /* Gibt einen zufaelligen Zustand aus, der dem Schritt hin zu Knoten
     * (a,b) entspricht, und zwar gemaess der bedingten Verteilung, dass
     * es von (a,b) aus mit einem Schritt von Zustand nextstep weitergeht.
     */
{
    a-=von_A+1; b-=von_B+1;
    if ( a<0 || b<0 ) 
	cerr << "Fehler: zu kleine Parameter in align_graph::zufallsschritt!!!\n";
    if (a<=0) {
	if (b<=0) cerr << "Fehler: a,b<=0 in  align_graph::zufallsschritt !!!\n";
	return A_GAP;
    }
    if (b==0) {
	return B_GAP;
    }
    double R=rand()/(RAND_MAX+1.0);
    double l_Paar=paar_kante[a-1][b-1]+fip.log_wkeit(PAARUNG,nextstep),
	l_A_GAP=A_gap_kante[a][b-1]+fip.log_wkeit(A_GAP,nextstep),
	l_B_GAP=B_gap_kante[a-1][b]+fip.log_wkeit(B_GAP,nextstep), 
	l_summe=addlog(l_Paar,l_A_GAP,l_B_GAP),
	Wk_A_gap=exp(l_A_GAP-l_summe),
	Wk_B_gap=exp(l_B_GAP-l_summe);
    if ( R < Wk_A_gap ) {
	return A_GAP;
    }
    if ( R < Wk_A_gap+Wk_B_gap ) {
	return B_GAP;
    }
    return PAARUNG;
}  

//++++++++++++++++++++++++++++++++++++++++++++
//    class maxwk_align_graph                +
//++++++++++++++++++++++++++++++++++++++++++++

// private:

void maxwk_align_graph::beschrifte(const alignment_mit_raten& algn) 
    /* Beschriftet die Kanten des align_graph 
   * gemaess den DNA-Staengen und den Parametern von algn
   */
{  
    const dna &dna_A=*(algn.get_dna_A());
    const dna &dna_B=*(algn.get_dna_B());
    sp=algn.subst_vorschlag.get_aktuell();
    fip=algn.indel_vorschlag.get_aktuell();
    // Zur W'keit fuer Substitution verwandte Logarithmen
    int i,j;
    A_gap_kante[0][0]=fip.log_wkeit(PAARUNG,A_GAP)+log(sp.get_basefreqs()[nukleo(dna_B[0])]);
    B_gap_kante[0][0]=fip.log_wkeit(PAARUNG,B_GAP)+log(sp.get_basefreqs()[nukleo(dna_A[0])]);
    paar_kante[0][0]=fip.log_wkeit(PAARUNG,PAARUNG)
	+ sp.log_wkeit(nukleo(dna_A[0]),nukleo(dna_B[0]));
    for(i=1;i<=dna_A.laen()-1;i++) {
	A_gap_kante[i][0]=B_gap_kante[i-1][0]
	    +fip.log_wkeit(B_GAP,A_GAP)
	    +log(sp.get_basefreqs()[nukleo(dna_B[0])]);
	B_gap_kante[i][0]=B_gap_kante[i-1][0]
	    +fip.log_wkeit(B_GAP,B_GAP)
	    +log(sp.get_basefreqs()[nukleo(dna_A[i])]);
	paar_kante[i][0]=B_gap_kante[i-1][0]
	    +fip.log_wkeit(B_GAP,PAARUNG)
	    +sp.log_wkeit(nukleo(dna_A[i]),nukleo(dna_B[0]));
    }
    A_gap_kante[dna_A.laen()][0]=B_gap_kante[dna_A.laen()-1][0]
	+fip.log_wkeit(B_GAP,A_GAP)+log(sp.get_basefreqs()[nukleo(dna_B[0])]);
    for(j=1;j<=dna_B.laen()-1;j++) {
	B_gap_kante[0][j]=A_gap_kante[0][j-1]
	    +fip.log_wkeit(A_GAP,B_GAP)
	    +log(sp.get_basefreqs()[nukleo(dna_A[0])]);
	A_gap_kante[0][j]=A_gap_kante[0][j-1]
	    +fip.log_wkeit(A_GAP,A_GAP)
	    +log(sp.get_basefreqs()[nukleo(dna_B[j])]);
	paar_kante[0][j]=A_gap_kante[0][j-1]
	    +fip.log_wkeit(A_GAP,PAARUNG)
	    + sp.log_wkeit(nukleo(dna_A[0]),nukleo(dna_B[j]));
    }
    B_gap_kante[0][dna_B.laen()]=A_gap_kante[0][dna_B.laen()-1]
				  +fip.log_wkeit(A_GAP,B_GAP)+log(sp.get_basefreqs()[nukleo(dna_A[0])]);
    for(i=1;i<=dna_A.laen()-1;i++) {
	for(j=1;j<=dna_B.laen()-1;j++) {
	    A_gap_kante[i][j]
		=maximum(B_gap_kante[i-1][j]+fip.log_wkeit(B_GAP,A_GAP)
			 +log(sp.get_basefreqs()[nukleo(dna_B[j])]),
			 paar_kante[i-1][j-1]+fip.log_wkeit(PAARUNG,A_GAP)
			 +log(sp.get_basefreqs()[nukleo(dna_B[j])]),
			 A_gap_kante[i][j-1]+fip.log_wkeit(A_GAP,A_GAP)
			 +log(sp.get_basefreqs()[nukleo(dna_B[j])]));
	    B_gap_kante[i][j]
		=maximum(B_gap_kante[i-1][j]+fip.log_wkeit(B_GAP,B_GAP)
			 +log(sp.get_basefreqs()[nukleo(dna_A[i])]),
			 paar_kante[i-1][j-1]+fip.log_wkeit(PAARUNG,B_GAP)
			 +log(sp.get_basefreqs()[nukleo(dna_A[i])]),
			 A_gap_kante[i][j-1]+fip.log_wkeit(A_GAP,B_GAP)
			 +log(sp.get_basefreqs()[nukleo(dna_A[i])]));
	    paar_kante[i][j]
		=maximum(B_gap_kante[i-1][j]+fip.log_wkeit(B_GAP,PAARUNG)
			 +sp.log_wkeit(nukleo(dna_A[i]),nukleo(dna_B[j])),
			 paar_kante[i-1][j-1]+fip.log_wkeit(PAARUNG,PAARUNG)
			 +sp.log_wkeit(nukleo(dna_A[i]),nukleo(dna_B[j])),
			 A_gap_kante[i][j-1]+fip.log_wkeit(A_GAP,PAARUNG)
			 +sp.log_wkeit(nukleo(dna_A[i]),nukleo(dna_B[j])));
	}
    }
    for(i=1;i<=dna_A.laen()-1;i++) 
	B_gap_kante[i][dna_B.laen()]
	    =maximum(B_gap_kante[i-1][dna_B.laen()]+fip.log_wkeit(B_GAP,B_GAP)
		     +log(sp.get_basefreqs()[nukleo(dna_A[i])]),
		     paar_kante[i-1][dna_B.laen()-1]+fip.log_wkeit(PAARUNG,B_GAP)
		     +log(sp.get_basefreqs()[nukleo(dna_A[i])]),
		     A_gap_kante[i][dna_B.laen()-1]+fip.log_wkeit(A_GAP,B_GAP)
		     +log(sp.get_basefreqs()[nukleo(dna_A[i])]));
    for(j=1;j<=dna_B.laen()-1;j++)
	A_gap_kante[dna_A.laen()][j]
	    =maximum(B_gap_kante[dna_A.laen()-1][j]+fip.log_wkeit(B_GAP,A_GAP)
		     +log(sp.get_basefreqs()[nukleo(dna_B[j])]),
		     paar_kante[dna_A.laen()-1][j-1]+fip.log_wkeit(PAARUNG,A_GAP)
		     +log(sp.get_basefreqs()[nukleo(dna_B[j])]),
		     A_gap_kante[dna_A.laen()][j-1]+fip.log_wkeit(A_GAP,A_GAP)
		     +log(sp.get_basefreqs()[nukleo(dna_B[j])]));
}

// public :

maxwk_align_graph::maxwk_align_graph(const alignment_mit_raten& algn)
    : align_graph(algn)
{
    /* Konstruiere einen Alignment-Graphen fuer das Alignment
     */
    beschrifte(algn);
}


zustand maxwk_align_graph::maxwk_schritt(zustand nextstep, int a, int b) const
    /* Gibt den wahrscheinlichsten Zustand aus, der dem Schritt hin zu Knoten
   * (a,b) entspricht, und zwar bezueglich der bedingten Verteilung, dass
   * es von (a,b) aus mit einem Schritt von Zustand nextstep weitergeht.
   */
{
    //cout << nextstep << endl;
    if ( a<0 || b<0 ) 
	cerr << "Fehler: zu kleine Parameter in align_graph::zufallsschritt!!!\n";
    if (a==0) {
	if (b==0) cerr << "Fehler: a=b=0 in  align_graph::zufallsschritt !!!\n";
	return A_GAP;
    }
    if (b==0) return B_GAP;
    double l_Paar=paar_kante[a-1][b-1]+fip.log_wkeit(PAARUNG,nextstep),
	l_A_GAP=A_gap_kante[a][b-1]+fip.log_wkeit(A_GAP,nextstep),
	l_B_GAP=B_gap_kante[a-1][b]+fip.log_wkeit(B_GAP,nextstep);
    if ( l_B_GAP < l_A_GAP ) { 
	if( l_Paar <  l_A_GAP )
	    return A_GAP;
	else
	    return PAARUNG;
    }
    else {
	if( l_Paar <  l_B_GAP )
	    return B_GAP;
	else
	    return PAARUNG;
    }
}  

