/****************************************************************************
 * This file is part of the program                                         *
 * mcmcsalut (Markov Chain Monte Carlo Sequence ALignment UTility) for      *
 * joint sampling alignments and mutation parameter values for two given    *
 * DNA-sequences.                                                           *
 *                                                                          *
 *    Copyright (C) 2000  Dirk Metzler                                      *
 *    e-mail: dmetzler@math.uni-frankfurt.de                                * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

class alignment_mit_raten;

//--------------------------------------------------------------------

class align_graph {
    friend ostream& operator<<(ostream& ostr, const align_graph& ag);
protected:
    int von_A, von_B, bis_A, bis_B; 
    substproc sp;
    fragment_indelproc fip;
    vector< vector<double> > A_gap_kante;
    vector< vector<double> > B_gap_kante;
    vector< vector<double> > paar_kante; 
public:
    align_graph(const alignment_mit_raten& algn,
		int von_A,int von_B,int bis_A,int bis_B)
	/* Konstruiere einen Alignment-Graphen fuer das Teilalignment zwischen
	 * (von_A,von_B) und (bis_A,bis_B) in algn mit den aktuellen Parametern
	 * von algn
	 */;
    align_graph(const alignment_mit_raten& algn)
	/* Konstruiere einen Alignment-Graphen fuer das Alignment
	 */;
    align_graph()
	/* Fuehrt zu Fehlermeldung
	 */;
};

//--------------------------------------------------------------------

class wk_align_graph : public  align_graph {
private:
    double ln_likelihood 
    /* loglikelihood der zur Konstruktion von wk_align_graph
     * verwendeten Parameterwerte, gegeben die Sequenzen 
     * (mit Summation ueber alle Alignments).
     */;
    void beschrifte(const alignment_mit_raten& algn) 
	/* Beschriftet die Kanten des maxwk_align_graph fuer den angegebenen 
	 * Bereich gemaess den DNA-Staengen und den Parametern von algn
	 */;
public:
    wk_align_graph(const alignment_mit_raten& algn,
		   int von_A,int von_B,int bis_A,int bis_B)
	/* Konstruiere einen maxwk_Alignment-Graphen fuer das Teilalignment
	 * zwischen
	 * (von_A,von_B) und (bis_A,bis_B) in algn mit den aktuellen Parametern
	 * von algn
	 */;
    double get_ln_likelihood() {return ln_likelihood;}
    zustand zufallsschritt(zustand s, int a, int b) const
	/* Gibt einen zufaelligen Zustand aus, der dem Schritt hin zu Knoten
	 * (a,b) entspricht, und zwar gemaess der bedingten Verteilung, dass
	 * es von (a,b) aus mit einem Schritt von Zustand s weitergeht.
	 */;
};

//--------------------------------------------------------------------

class maxwk_align_graph : public  align_graph {
private: 
    void beschrifte(const alignment_mit_raten& algn) 
	/* Beschriftet die Kanten des maxwk_align_graph fuer den angegebenen 
	 * Bereich gemaess den DNA-Staengen und den Parametern von algn
	 */;
public:
    maxwk_align_graph(const alignment_mit_raten& algn)
	/* Konstruiere einen maxwk_Alignment-Graphen fuer das Alignment
	 */;
    zustand maxwk_schritt(zustand s, int a, int b) const
	/* Gibt den wahrscheilichsten Zustand aus, der dem Schritt hin zu Knoten
	 * (a,b) entspricht, und zwar bezueglich der bedingten Verteilung, dass
	 * es von (a,b) aus mit einem Schritt von Zustand s weitergeht.
	 */;
};

ostream& operator<<(ostream& ostr, const align_graph& ag);

//--------------------------------------------------------------------

class alignment_mit_raten {
    friend class align_graph;
    friend class wk_align_graph;
    friend class maxwk_align_graph;
    friend ostream& operator<<(ostream& ostr, const alignment_mit_raten& algn);
private:
    const dna *dna_A, *dna_B;
    fragindel_proposal_chain indel_vorschlag;
    substproc_proposal_chain subst_vorschlag;
    bool fix_subst_indel_ratio;
    vector<int> dAB,dBA;
    void set_max_wk()
	/* Setzt das Alignment auf das wahrscheinlichste fuer die aktuellen 
	 * Parameter
	 */;
    void zaehle_ueberg(int anz_ueberg[4][5],
		       int anz_paare[4][4],
		       int anz_indel_basen[4],
		       int von_A,int von_B,int bis_A,int bis_B) const 
	/* Berechnet die Anzahlen der verschiedenen Uebergaenge zwischen den
	 * Zustaenden, die Anzahl der verschiedenen Basenpaarungen sowie der 
	 * nichtalignierten Basen im aktuellen Alignment im angegebenen 
	 * Bereich und schreibt die Ergebnisse in die uebergebenen Felder.
	 */;
    zustand get_zust(int a,int b) const
	/* Liefert den zu a und b assoziierten Alignment-Zustand 
	 * (als inline definiert)
	 */; 
    zustand iter_zust(int* pa,int* pb) const
	/* Liefert den zu *pa und *pb assoziierten Alignment-Zustand und 
	 * setzt (*pa,*pb) einen Zustand weiter.
	 */;
    void teilsample(int von_A, int von_B, int bis_A, int bis_B) 
	/* sample ein neues Teilalignment von (von_A,von_B) 
	 * bis (bis_A,bis_B). Dabei bleiben die Zustaende der von
	 * Start- und Endpunkt erhalten.
	 */;
    double get_ln_wk(substproc sp, fragment_indelproc fip) const 
	/* liefert den ln der Wkeit des aktuellen Alignments fuer 
	 * die uebergebenen Parameter.
	 */;
    alignment_mit_raten()
	/* Wird nicht definiert; erzeugt daher Compiler-Fehlermeldung.
	 */;   
    void coupled_param_resamp()
	/* Fuehrt einen Metropolis-Hastings-Schritt fuer die Mutationsraten
	 * bedingt auf das aktuelle Alignment durch. Dabei bleiben die Verhaeltnisse
	 * zwischen den verschiedenen Raten erhalten. Es entspricht also einem
	 * samplen der Zeitspanne zwischen den Sequenzen.
	 */;
    void indel_param_resamp() 
	/* Fuehrt einen Metropolis-Hastings-Schritt fuer die indel-parameter
	 * bedingt auf das aktuelle Alignment und aktuelle substparameter durch.
	 */;
    void subst_param_resamp()
	/* Fuehrt einen Metropolis-Hastings-Schritt fuer die subst-parameter
	 * bedingt auf das aktuelle Alignment und aktuelle indelparameter durch.
	 */;
public:
    alignment_mit_raten(const dna &d_1,const dna &d_2,
	      const dna_with_gaps &d_1wg,const dna_with_gaps &d_2wg,
	      const eingabe_parameter & ep)
	/* Konstruiert das Alignment gemaess den Eingabeparametern in eg
	 */;
    const dna* get_dna_A() const {return dna_A;};
    const dna* get_dna_B() const {return dna_B;};
    void sample()
	/* sample das Alignment
	 */;
    void teilsample(int von_A, int bis_A)
	/* sample ein neues Teilalignment von dna_A[von_A] 
	 * bis dna_A[bis_A].
	 */;
    void teilsample(int laenge)
	/* sample ein neues Teilalignment zu einem zufaelligen Abschnitt der 
	 * Laenge laenge. 
	 */;
    void param_resamp()
	/* Fuehrt einen Metropolis-Hastings-Schritt fuer s
	 * lambda und mu bedingt auf das aktuelle Alignment durch.
	 */;  
    void print_params(ostream& ostr)
	/* gibt s, lambda und ln(W'keit) in den ostr aus
	 */;
    void print_path(ostream& ostr) 
	/* gibt das Alignment Pfad in ostr aus:
	 * d,v,h stehen fuer diagonal, vertikal und horizontal
	 */;
    double get_ln_wk() const
	/* liefert den ln der Wkeit des aktuellen Alignments fuer die 
	 * aktuellen Parameter. (als inline definiert)
	 */;
};

inline zustand alignment_mit_raten::get_zust(int a,int b) const {
    /* Liefert den zu a und b assoziierten Alignment-Zustand
     */
    if (a==-1) {
	if (b==-1) 
	    return START;
	if (dBA[b]==-1) 
	    return A_GAP;
	return NOSTATE;
    }
    if (a==dna_A->laen()) {
	if (b==dna_B->laen())
	    return ENDE;
	return NOSTATE;
    }
    if (b==-1) {
	if (dAB[a]==-1) 
	    return B_GAP;
	return NOSTATE;
    }
    if (b==dna_B->laen()) {
	return NOSTATE;
    }
    if (b==dAB[a]){
	if (a==dBA[b])
	    return PAARUNG;
	if (dBA[b]<a)
	    return B_GAP;
	return NOSTATE;
    }
    if (dBA[b]==a && dAB[a]<b )
	return A_GAP;
    return NOSTATE;
}

inline double alignment_mit_raten::get_ln_wk() const {
    /* liefert den ln der Wkeit des aktuellen Alignments fuer die 
     * aktuellen Parameter.
     */
    return get_ln_wk(subst_vorschlag.get_aktuell(),indel_vorschlag.get_aktuell());
}

ostream& operator<<(ostream& ostr, const alignment_mit_raten& algn);
