/****************************************************************************
 * This file is part of the program                                         *
 * mcmcsalut (Markov Chain Monte Carlo Sequence ALignment UTility) for      *
 * joint sampling alignments and mutation parameter values for two given    *
 * DNA-sequences.                                                           *
 *                                                                          *
 *    Copyright (C) 2000  Dirk Metzler                                      *
 *    e-mail: dmetzler@math.uni-frankfurt.de                                * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <cmath>
#include <vector>
#include <utility>
#include <iostream>
#include <iomanip>

#include "indelproc.hh"

//  +++++++++++++++++++++++++++++++
//  +   class fragment_indelproc  +
//  +++++++++++++++++++++++++++++++

// private:

void fragment_indelproc::update_log_ueb_wkeiten() {
    if(lambda<1/1000000.0) {
	log_ueb_wkeiten[PAARUNG][PAARUNG]=log(1-(2*lambda)/(gamma*(1+lambda)));
	log_ueb_wkeiten[PAARUNG][A_GAP]=log(lambda/(gamma*(1+lambda)));
	log_ueb_wkeiten[PAARUNG][B_GAP]=log(lambda/(gamma*(1+lambda)));
	log_ueb_wkeiten[PAARUNG][ENDE]=log_ueb_wkeiten[PAARUNG][PAARUNG];
	log_ueb_wkeiten[A_GAP][PAARUNG]=log((1-lambda)/(gamma*(1+lambda)));
	log_ueb_wkeiten[A_GAP][A_GAP]=log((gamma*(1+lambda)-1)/(gamma*(1+lambda)));
	log_ueb_wkeiten[A_GAP][B_GAP]=log(lambda/(gamma*(1+lambda)));
	log_ueb_wkeiten[A_GAP][ENDE]=log_ueb_wkeiten[A_GAP][PAARUNG];
	log_ueb_wkeiten[B_GAP][PAARUNG]=log(lambda*(1-lambda)
				      /(gamma*(lambda)*(1+lambda)));
	log_ueb_wkeiten[B_GAP][A_GAP]=log(lambda/(gamma*(1+lambda)));
	log_ueb_wkeiten[B_GAP][B_GAP]=log(1-1/(gamma*(1+lambda)));
	log_ueb_wkeiten[B_GAP][ENDE]=log_ueb_wkeiten[B_GAP][PAARUNG];
	log_ueb_wkeiten[START][PAARUNG]=log_ueb_wkeiten[PAARUNG][PAARUNG];
	log_ueb_wkeiten[START][A_GAP]=log_ueb_wkeiten[PAARUNG][A_GAP];
	log_ueb_wkeiten[START][B_GAP]=log_ueb_wkeiten[PAARUNG][B_GAP];
	log_ueb_wkeiten[START][ENDE]=log_ueb_wkeiten[PAARUNG][PAARUNG];
    }
    else {
	log_ueb_wkeiten[PAARUNG][PAARUNG]=log(1-(1+lambda-exp(-lambda))
					/(gamma*(1+lambda)));
	log_ueb_wkeiten[PAARUNG][A_GAP]=log(lambda/(gamma*(1+lambda)));
	log_ueb_wkeiten[PAARUNG][B_GAP]=log((1-exp(-lambda))/(gamma*(1+lambda)));
	log_ueb_wkeiten[PAARUNG][ENDE]=log_ueb_wkeiten[PAARUNG][PAARUNG];
	log_ueb_wkeiten[A_GAP][PAARUNG]=log(exp(-lambda)/(gamma*(1+lambda)));
	log_ueb_wkeiten[A_GAP][A_GAP]=log((gamma*(1+lambda)-1)/(gamma*(1+lambda)));
	log_ueb_wkeiten[A_GAP][B_GAP]=log((1-exp(-lambda))/(gamma*(1+lambda)));
	log_ueb_wkeiten[A_GAP][ENDE]=log_ueb_wkeiten[A_GAP][PAARUNG];
	log_ueb_wkeiten[B_GAP][PAARUNG]=log(lambda*exp(-lambda)
				      /(gamma*(1-exp(-lambda))*(1+lambda)));
	log_ueb_wkeiten[B_GAP][A_GAP]=log((1-exp(-lambda)*(1+lambda))
				    /(gamma*(1-exp(-lambda))*(1+lambda)));
	log_ueb_wkeiten[B_GAP][B_GAP]=log(1-1/(gamma*(1+lambda)));
	log_ueb_wkeiten[B_GAP][ENDE]=log_ueb_wkeiten[B_GAP][PAARUNG];
	log_ueb_wkeiten[START][PAARUNG]=log_ueb_wkeiten[PAARUNG][PAARUNG];
	log_ueb_wkeiten[START][A_GAP]=log_ueb_wkeiten[PAARUNG][A_GAP];
	log_ueb_wkeiten[START][B_GAP]=log_ueb_wkeiten[PAARUNG][B_GAP];
	log_ueb_wkeiten[START][ENDE]=log_ueb_wkeiten[PAARUNG][PAARUNG];
    }
//     cout << "fragment_indelproc::update_log_ueb_wkeiten()" << endl;
//     for (int i=0; i<4; i++) {
//       for (int j=0; j<5; j++) 
// 	cout << setw(15) << exp(log_ueb_wkeiten[i][j]);
//       cout << endl;
//     };
}

// public:

fragment_indelproc::fragment_indelproc(double lam, double gam) {
    set(lam,gam);
    update_log_ueb_wkeiten();
};

fragment_indelproc::fragment_indelproc(const fragment_indelproc& fip) {
    *this=fip;
};

fragment_indelproc& 
fragment_indelproc::operator=(const fragment_indelproc& fip){
    lambda=fip.lambda;
    gamma=fip.gamma;
    for (int i=0; i<4; i++)
	for (int j=0; j<5; j++)
	    log_ueb_wkeiten[i][j]=fip.log_ueb_wkeiten[i][j];
    return *this;
};

ostream& operator<<(ostream& ostr, const fragment_indelproc& fip) {
    return ostr << "lambda=" << fip.get_lambda() 
		<< " gamma=" << fip.get_gamma() << endl;
}
