/****************************************************************************
 * This file is part of the program                                         *
 * mcmcsalut (Markov Chain Monte Carlo Sequence ALignment UTility)          *
 * for joint sampling alignments and mutation parameter values for two      *
 * given DNA-sequences.                                                     *
 *                                                                          *
 *    Copyright (C) 2000  Dirk Metzler                                      *
 *    e-mail: dmetzler@math.uni-frankfurt.de                                * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/
#include <iostream>
#include <vector>
#include <cmath>
#include <fstream>
#include <string>

using namespace std;

#include "substproc.hh"
#include "indelproc.hh"
#include "propose.hh"

double rand_exp(double x) {
    /* gibt exp(x)-verteilte ZVE zurueck.
     * (exp-verteilt mit Erwartungswert 1/x)
     */
    double R=rand()/(RAND_MAX+1.0);
    if ( R>0.99 ) return -log(R)/x;
    else return -log(0.99)/x+rand_exp(x);
}

inline double rand_quot(double q) {
    /* veraendert den quotienten q=x/y so, dass im Gleichgeicht 
     * x/(x+y) uniform in [0,1] ist. Dabei wird [0,1] zyklisch geschlossen.
     */
    double x=q/(1+q), r=x+rand()/(RAND_MAX*10.0)-1.0/20.0;
    x = (r<0 ? 1+r : (r>=1 ? r-1 : r));
    return  x/(1-x);
}

inline double quotient(double x, double y) {
  if (y==0) {
    if (x!=0) {
      cerr << "x/y=infinite in inline double quotient(double x, double y)\n";
      exit(1);
    }
    return 1.0;
  }
  return x/y;
}

//----------------class substproc_proposal_chain-------------------------

// public:

// public:

substproc_proposal_chain::substproc_proposal_chain(const substproc & sp,
						   const vector<bool> & fix) 
    : aktuell(sp) {

    fix_GC_cont=fix[0];
    fix_A_zu_T=fix[1];
    fix_C_zu_G=fix[2];
    fix_allgrate=fix[3]; 
    fix_Ti_zu_Tv=fix[4];
    fix_Pu_zu_Py=fix[5];
    vorlaeufig=aktuell;
}

double substproc_proposal_chain::quot_prior() const {
    /* liefert den Quotienten zwischen der Dichte der a-priori-Verteilung
     * bezgl. der vorlaeufigen und der aktuellen Parameter
     */
  //   cout << vorlaeufig.get_allgrate() << ' ' << aktuell.get_allgrate() << endl
  // << vorlaeufig.get_zusAG() << ' ' << aktuell.get_zusAG() << endl
  // << vorlaeufig.get_zusCT() << ' ' << aktuell.get_zusCT() << endl;
    return exp( - ( vorlaeufig.get_allgrate() - aktuell.get_allgrate()
		    + vorlaeufig.get_zusAG()  - aktuell.get_zusAG()
		    + vorlaeufig.get_zusCT()  - aktuell.get_zusCT() ) 
	       );
}

double substproc_proposal_chain::uebergangs_quotient() const {
    /* liefert p(vorlaeufig->aktuell)/p(aktuell->vorlaeufig)
     * (wichtig fuer Metropolis-Hastings)
     */
    double AT_zu_GC=quotient(aktuell.get_basefreqs()[A]+aktuell.get_basefreqs()[T]
		     ,aktuell.get_basefreqs()[G]+aktuell.get_basefreqs()[C]),
	allgrate=aktuell.get_allgrate(),
	Ti_zu_allg=quotient(aktuell.get_zusAG()+aktuell.get_zusCT(),aktuell.get_allgrate()),
	vAT_zu_GC=quotient(vorlaeufig.get_basefreqs()[A]+vorlaeufig.get_basefreqs()[T]
		     ,vorlaeufig.get_basefreqs()[G]+vorlaeufig.get_basefreqs()[C]),
	vallgrate=vorlaeufig.get_allgrate(),
	vTi_zu_allg=quotient(vorlaeufig.get_zusAG()+vorlaeufig.get_zusCT()
			     ,vorlaeufig.get_allgrate());
    return quotient(AT_zu_GC*allgrate*Ti_zu_allg
		    ,vAT_zu_GC*vallgrate*vTi_zu_allg)
	*exp( quotient(vAT_zu_GC,AT_zu_GC) - quotient(AT_zu_GC,vAT_zu_GC)
	      + quotient(vallgrate,allgrate) - quotient(allgrate,vallgrate)
	      + quotient(vTi_zu_allg,Ti_zu_allg) - quotient(Ti_zu_allg,vTi_zu_allg));
}

void substproc_proposal_chain::vorschlag() { 
    /* findet ein einen neuen Kandidaten fuer vorlaeufig
     */
    double AT_zu_GC=((aktuell.get_basefreqs()[A]+aktuell.get_basefreqs()[T])
		     /(aktuell.get_basefreqs()[G]+aktuell.get_basefreqs()[C])),
	A_zu_T=aktuell.get_basefreqs()[A]/aktuell.get_basefreqs()[T],
	C_zu_G=aktuell.get_basefreqs()[C]/aktuell.get_basefreqs()[G],
	allgrate=aktuell.get_allgrate(),
	Ti_zu_allg=(aktuell.get_zusAG()+aktuell.get_zusCT())/aktuell.get_allgrate(),
	Pu_zu_Py= (aktuell.get_zusAG()==aktuell.get_zusCT() ? 1.0 :
		   (aktuell.get_zusCT()<=0.0 ? 100000.0 : aktuell.get_zusAG()/aktuell.get_zusCT()));
//     cout << endl << "substproc_proposal_chain::vorschlag()" << endl
// 	 << "AT_zu_GC=" << AT_zu_GC << ",A_zu_T="<< A_zu_T << ",C_zu_G="<< C_zu_G << ",allgrate="
// 	 << allgrate << ",Ti_zu_allg=" << Ti_zu_allg << ",Pu_zu_Py=" <<Pu_zu_Py << endl;
    if(!fix_GC_cont && AT_zu_GC>0.0 ) AT_zu_GC=rand_exp(1/AT_zu_GC);
    if(!fix_A_zu_T && A_zu_T>0.0 ) A_zu_T=rand_quot(A_zu_T);
    if(!fix_C_zu_G && C_zu_G>0.0 ) C_zu_G=rand_quot(C_zu_G);
    if(!fix_allgrate && allgrate>0.0 )  allgrate=rand_exp(1/allgrate);
    if(!fix_Ti_zu_Tv && Ti_zu_allg>0.0 ) Ti_zu_allg=rand_exp(1/Ti_zu_allg);
    if(!fix_Pu_zu_Py && Pu_zu_Py>0.0 )  Pu_zu_Py=rand_quot(Pu_zu_Py);
    double AplusT = AT_zu_GC/(1+AT_zu_GC);
    double freq[4];
    freq[A]=AplusT*A_zu_T/(1+A_zu_T);
    freq[C]=(1-AplusT)*C_zu_G/(1+C_zu_G);
    freq[G]=1-AplusT-freq[C];
    freq[T]=AplusT-freq[A];
    vorlaeufig.set( freq , allgrate , Ti_zu_allg*Pu_zu_Py/(1+Pu_zu_Py)*allgrate
		    , Ti_zu_allg*(1-Pu_zu_Py/(1+Pu_zu_Py))*allgrate );
//     cout << "AT_zu_GC=" << AT_zu_GC << ",A_zu_T="<< A_zu_T << ",C_zu_G="<< C_zu_G << ",allgrate="
// 	 << allgrate << ",Ti_zu_allg=" << Ti_zu_allg << ",Pu_zu_Py=" <<Pu_zu_Py << endl
// 	 << "substproc_proposal_chain::vorschlag()" << endl
// 	 << "aktuell    : " << aktuell << endl 
// 	 << "vorlaeufig : " << vorlaeufig << endl;
}

void substproc_proposal_chain::vorschlag(double t) {
    /* als Kandidaten fuer vorlaeufig werden die aktuellen
     * Substitutionsraten um den Faktor t
     * gestreckt.
     */    
    double freq[4]={aktuell.get_basefreqs()[0],aktuell.get_basefreqs()[1],
		    aktuell.get_basefreqs()[2],aktuell.get_basefreqs()[3]};
    vorlaeufig.set(freq,aktuell.get_allgrate()*t,aktuell.get_zusAG()*t,aktuell.get_zusCT()*t);
}

void substproc_proposal_chain::akzeptiere() {
    /* waehlt die vorlaeufigen Parameter als aktuelle
     */
    aktuell=vorlaeufig;
}

// ---------------- class fragindel_proposal_chain -------------------

// private:

// public:

fragindel_proposal_chain::fragindel_proposal_chain(const fragment_indelproc & fip, 
						   const vector<bool> & fix, 
						   double priorgamma) 
    : aktuell(fip), vorlaeufig(fip) {
    fix_lambda=fix[0];
    fix_gamma=fix[1];
    prior_gamma_erw=priorgamma;
}

double fragindel_proposal_chain::quot_prior() const {
    /* liefert den Quotienten zwischen der Dichte der a-priori-Verteilung
     * bezgl. der vorlaeufigen und der aktuellen Parameter
     */
  // cout <<  vorlaeufig.get_lambda() << ' ' << aktuell.get_lambda() << endl
  // << vorlaeufig.get_gamma() << ' ' << aktuell.get_gamma() << endl
  // << "..................................................\n";
    return exp( - ( vorlaeufig.get_lambda() - aktuell.get_lambda() )
		- ( vorlaeufig.get_gamma() - aktuell.get_gamma() ) 
		/ prior_gamma_erw
		);
}

double fragindel_proposal_chain::uebergangs_quotient() const {
    /* liefert p(vorlaeufig->aktuell)/p(aktuell->vorlaeufig)
     * (wichtig fuer Metropolis-Hastings)
     */
    double lambda=aktuell.get_lambda(), new_lambda=vorlaeufig.get_lambda(),
	gamma=aktuell.get_gamma(), new_gamma=vorlaeufig.get_gamma();
    return lambda/new_lambda*exp(new_lambda/lambda-lambda/new_lambda)
	*gamma/new_gamma*exp((new_gamma-1)/gamma-(gamma-1)/new_gamma);
}

void fragindel_proposal_chain::vorschlag() { 
    /* findet einen Kandidaten fuer vorlaeufig
     */
    vorlaeufig.set(fix_lambda==1 ? aktuell.get_lambda() 
		   : rand_exp(1/aktuell.get_lambda()),
		   fix_gamma==1 ? aktuell.get_gamma() 
		   : 1.0+rand_exp(1/aktuell.get_gamma()));
}

void fragindel_proposal_chain::vorschlag(double t) {
    /* als Kandidaten fuer vorlaeufig wird lambda um den Faktor
     * gestreckt.
     */;
     vorlaeufig.set(aktuell.get_lambda()*t,aktuell.get_gamma());
}

void fragindel_proposal_chain::akzeptiere() {
    /* waehlt die vorlaeufigen Parameter als aktuelle
     */
    aktuell=vorlaeufig;
}
