/****************************************************************************
 * This file is part of the program                                         *
 * mcmcsalut (Markov Chain Monte Carlo Sequence ALignment UTility)  for     *
 * joint sampling alignments and mutation parameter values for two given    *
 * DNA-sequences.                                                           *
 *                                                                          *
 *    Copyright (C) 2000  Dirk Metzler                                      *
 *    e-mail: dmetzler@math.uni-frankfurt.de                                * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <vector>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <cmath>

#include "substproc.hh"

//+++++++++++++++++++++++++++++++++++++++++++
//             class dna_with_gaps          +
//+++++++++++++++++++++++++++++++++++++++++++

// private:

// public:

dna_with_gaps::dna_with_gaps(){
} 

dna_with_gaps::dna_with_gaps(ifstream& datei)
{
    char B;
    while(datei.get(B)) {
	switch (B) {
	case 'A':
	case 'a':
	    push_back('A');
	    break;
	case 'C':
	case 'c':
	    push_back('C');
	    break;
	case 'G':
	case 'g':
	    push_back('G');
	    break;
	case 'T':
	case 't':
	    push_back('T');
	    break;
	case '_':
	case '-':
	    push_back('_');
	    break;
	case '\n':
	case ' ':
	    break;
	default:
	    cerr << "Fehler in Eingabe-Datei: "<< B 
		 << " Bitte nur A,C,G,T !" << endl;
	}
    }
}

dna_with_gaps::dna_with_gaps(string infile) {
    
    ifstream datei(infile.c_str());
    char B;
    if (!datei) {
      cerr << "cannot open file " << infile << endl;
      exit(EXIT_FAILURE);
    }
    while(datei.get(B)) {
	switch (B) {
	case 'A':
	case 'a':
	    push_back('A');
	    break;
	case 'C':
	case 'c':
	    push_back('C');
	    break;
	case 'G':
	case 'g':
	    push_back('G');
	    break;
	case 'T':
	case 't':
	    push_back('T');
	    break;
	case '_':
	case '-':
	    push_back('_');
	    break;
	case '\n':
	case ' ':
	    break;
	default:
	    cerr << "Fehler in Eingabe-Datei: "<< B 
		 << " Bitte nur A,C,G,T !" << endl;
	}
    }
    datei.close();   
}

// dna_with_gaps::~dna_with_gaps(){}

ostream& operator<<(ostream& ostr,const dna_with_gaps& d) {
    for (int i=0; i<d.laen(); ++i){
	ostr << d[i];
    }
    return ostr;
}

//+++++++++++++++++++++++++++++++++++++++++++
//             class dna                    +
//+++++++++++++++++++++++++++++++++++++++++++

// private:

// public:

dna::dna(ifstream& datei)
{
    char B;
    while(datei.get(B)) {
	switch (B) {
	case 'A':
	case 'C':
	case 'G':
	case 'T':
	    push_back(B);
	    break;
	default:
	    cerr << "error in input file: "<< B 
		 << " only A,C,G, and T are allowed!" << endl;
	}
    }
}

dna::dna(const dna_with_gaps &dwg)
{
    for(int i=0; i<dwg.laen(); i++) {
	switch (dwg[i]) {
	case '_':
	    break;
	case 'A':
	case 'C':
	case 'G':
	case 'T':
	    push_back(dwg[i]);
	    break;
	default:
	    cerr << "Fehler beim Konstruieren einer DNA: Base " 
		 << dwg[i] << endl;
	}
    }
}


//+++++++++++++++++++++++++++++++++++++++++++
//             class substproc              +
//+++++++++++++++++++++++++++++++++++++++++++


void substproc::update_log_paar_wkeiten() {
    log_paar_wkeit[A][A]= log( basefreq[A] *
			       ( // keine Mutation:
				exp(-allgrate-zusAG)
				// mit Mutation:
				+  // keine allgemeine Mutation
				exp(-allgrate)*(1-exp(-zusAG))
				* basefreq[A]/(basefreq[A]+basefreq[G])
				+ // mit allgemeiner Mutation
				(1-exp(-allgrate))*basefreq[A]
				)
			       );
    log_paar_wkeit[A][C]= log( basefreq[A] * (1-exp(-allgrate)) * basefreq[C] );
    log_paar_wkeit[A][G]= log( basefreq[A] *
			       ( // keine allgemeine Mutation
				exp(-allgrate)*(1-exp(-zusAG))
				* basefreq[G]/(basefreq[A]+basefreq[G])
				+ // mit allgemeiner Mutation
				(1-exp(-allgrate))*basefreq[G]
				)
			       );
    log_paar_wkeit[A][T]= log( basefreq[A] * (1-exp(-allgrate)) * basefreq[T] );
    log_paar_wkeit[C][A]= log_paar_wkeit[A][C]; 
    log_paar_wkeit[C][C]= log( basefreq[C] *
			       ( // keine Mutation:
				exp(-allgrate-zusCT)
				// mit Mutation:
				+  // keine allgemeine Mutation
				exp(-allgrate)*(1-exp(-zusCT))
				* basefreq[C]/(basefreq[C]+basefreq[T])
				+ // mit allgemeiner Mutation
				(1-exp(-allgrate))*basefreq[C]
				)
			       );
    log_paar_wkeit[C][G]= log( basefreq[C] * (1-exp(-allgrate)) * basefreq[G] );
    log_paar_wkeit[C][T]= log( basefreq[C] *
			       ( // keine allgemeine Mutation
				exp(-allgrate)*(1-exp(-zusCT))
				* basefreq[T]/(basefreq[C]+basefreq[T])
				+ // mit allgemeiner Mutation
				(1-exp(-allgrate))*basefreq[T]
				)
			       );
    log_paar_wkeit[G][A]= log_paar_wkeit[A][G];
    log_paar_wkeit[G][C]= log_paar_wkeit[C][G];
    log_paar_wkeit[G][G]= log( basefreq[G] *
			       ( // keine Mutation:
				exp(-allgrate-zusAG)
				// mit Mutation:
				+  // keine allgemeine Mutation
				exp(-allgrate)*(1-exp(-zusAG))
				* basefreq[G]/(basefreq[A]+basefreq[G])
				+ // mit allgemeiner Mutation
				(1-exp(-allgrate))*basefreq[G]
				)
			       );
    log_paar_wkeit[G][T]= log( basefreq[G] * (1-exp(-allgrate)) * basefreq[T] );
    log_paar_wkeit[T][A]= log_paar_wkeit[A][T];
    log_paar_wkeit[T][C]= log_paar_wkeit[C][T];
    log_paar_wkeit[T][G]= log_paar_wkeit[G][T];
    log_paar_wkeit[T][T]= log( basefreq[T] *
			       ( // keine Mutation:
				exp(-allgrate-zusCT)
				// mit Mutation:
				+  // keine allgemeine Mutation
				exp(-allgrate)*(1-exp(-zusCT))
				* basefreq[T]/(basefreq[C]+basefreq[T])
				+ // mit allgemeiner Mutation
				(1-exp(-allgrate))*basefreq[T]
				)
			       );
//     cout << "substproc::update_log_paar_wkeiten()" << endl;   
//     for( int i=0; i<4; i++){
// 	for (int j=0; j<4; j++) 
// 	    cout << setw(15) << exp(log_paar_wkeit[i][j]);
// 	cout << endl;
//     }
}

// public:

substproc::substproc() {
    /* setzt die allgemeine Substitutionsrate auf 0.1
     * und den Rest auf Defaultwerte
     */
    basefreq[A]=0.25;
    basefreq[C]=0.25;
    basefreq[G]=0.25;
    basefreq[T]=0.25;
    allgrate=0.1;
    zusAG=0.01; 
    zusCT=0.01;
    update_log_paar_wkeiten();
}

substproc::substproc(const double freq[4], double s, double zAG, double zCT) {
    /* setzt die allgemeine Substitutionsrate auf s, die 
     * Basenhaeufigkeiten auf freq und Raten fuer zusaetzliche
     * Purin- und Pyrimidinsubstitutionen auf zAG und zCT.
     */
    set(freq,s,zAG,zCT);
}

substproc::substproc(const substproc& sp){
    *this=sp;
}; 

substproc& substproc::operator=(const substproc& sp){
    for (int i=0; i<4; i++) {
	basefreq[i]=sp.basefreq[i];
	for (int j=0; j<4; j++) 
	    log_paar_wkeit[i][j]=sp.log_paar_wkeit[i][j];
    }
    allgrate=sp.allgrate; 
    zusAG=sp.zusAG;
    zusCT=sp.zusCT;
    return *this;
}; 

void substproc::set(const double freq[4], double allg, double AG,double CT) {
    /* Die Funktion set(...) setzt die entsprechenden
     * Parameter und updatet log_paar_wk
     */
    for (int i=0; i<4; ++i) basefreq[i]=freq[i];
    allgrate=allg; zusAG=AG; zusCT=CT;
    update_log_paar_wkeiten();
}

ostream& operator<<(ostream& ostr, const substproc& sp) {
    return ostr << "p(A)=" << sp.get_basefreqs()[A]
		<< ", p(C)=" << sp.get_basefreqs()[C]
		<< ", p(G)=" << sp.get_basefreqs()[G]
		<< ", p(T)=" << sp.get_basefreqs()[T] << endl
		<< "substitution rate=" << sp.get_allgrate()
		<< ", additional A<->G rate=" << sp.get_zusAG()
		<< ", additional C<->T rate=" << sp.get_zusCT();
} 
