/****************************************************************************
 * This file is part of the program                                         *
 * mcmcsalut (Markov Chain Monte Carlo Sequence ALignment UTility) for      *
 * joint sampling alignments and mutation parameter values for two given    *
 * DNA-sequences.                                                           *
 *                                                                          *
 *    Copyright (C) 2000  Dirk Metzler                                      *
 *    e-mail: dmetzler@math.uni-frankfurt.de                                * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

enum nukleotid {A=0, C=1, G=2, T=3};

inline nukleotid nukleo(const char& c){
    // wandelt char in nukleotid um
    switch (c) {
    case 'A' :
    case 'a' :
	return A;
    case 'C' :
    case 'c' :
	return C;
    case 'G' :
    case 'g' :
	return G;
    case 'T' :
    case 't' :
	return T;
    default :
	cerr << "unbekantes Nukleotid \"" << c << "\" in nukleo(const char& c)\n";
	exit(1);
    }
}

class dna_with_gaps : public vector <char> {
public:
    dna_with_gaps();
    dna_with_gaps(ifstream& datei);
    dna_with_gaps(string infile);
    // ~dna_with_gaps();
    nukleotid site(int n) const {
	// gibt das Nukleotid an site n zurueck
	if (n >= laen()) {
	    cerr << "n >= laen() in nukleotid site(int n)\n";
	    exit(1);
	}
	return nukleo((*this)[n]);
    }
    int laen() const {
	// gibt laenge zurueck
	return size();
    }
};

class dna : public dna_with_gaps {
public:
    dna(ifstream& datei);
    dna(const dna_with_gaps &dwg);
};

ostream& operator<<(ostream& ostr, const dna_with_gaps& d);

class substproc {
private:
    double basefreq[4], // Basenhauefigkeiten
	log_paar_wkeit[4][4], // logarithmen der wkeiten der Paarungen
	allgrate, // allgemeine Substitutionsrate (auch nicht-veraendernd)
	zusAG, // rate der zusaetzlichen Purinsubstitutionen (-"-)
	zusCT; // rate der zusaetzlichen Pyrimidinsubstitutionen (-"-)
    void update_log_paar_wkeiten();
public:
    substproc()
	/* setzt die allgemeine Substitutionsrate auf 0.1
	 * und den Rest auf Defaultwerte
	 */;
    substproc(const double freq[4], double s, double zAG, double zCT)
	/* setzt die allgemeine Substitutionsrate auf s, die 
	 * Basen haeufigkeiten auf freq und Raten fuer zusaetzliche
	 * Purin- und Pyrimidinsubstitutionen auf zAG und zCT.
	 */;
    substproc(const substproc&); 
    substproc& operator=(const substproc&); 
    void set(const double freq[4], double allg, double AG,double CT)
	/* Die Funktion set(...) setzt die entsprechenden
	 * Parameter und updatet log_paar_wk
	 */;
    // Die Funktionen get..(..) liefern den Wert des jeweiligen Parameters
    vector<double> get_basefreqs() const {
	vector<double> freqs(4);
	for (int i=0; i<4; ++i) freqs[i]=basefreq[i];
	return freqs; 
    }
    double get_allgrate() const { return allgrate; }
    double get_zusAG() const { return zusAG; }
    double get_zusCT() const { return zusCT; }  
    double log_wkeit(nukleotid a, nukleotid b) const { return log_paar_wkeit[a][b]; }
};



ostream& operator<<(ostream& ostr, const substproc& sp);
